
#include <stdio.h>
#include "pvm3.h"


#include "life.h"


void main() {
  int tid_host, tids [NNODES] ;
  int node_id, node_prev, node_next ;
  int dummy ;
  int blockSizeMax, blockBase, blockSize ;
  int (*block) [N], (*neighbours) [N] ;
  int i, y ;
  int iter ;

  pvm_mytid() ;

  tid_host = pvm_parent() ;

  /*  Get `node_id' and `tids' vector from host.  */

  pvm_precv(tid_host, 0, &node_id,  1, PVM_INT, &dummy, &dummy, &dummy) ;
  pvm_precv(tid_host, 0, tids, NNODES, PVM_INT, &dummy, &dummy, &dummy) ;

  node_prev = (node_id - 1 + NNODES) % NNODES ;
  node_next = (node_id + 1) % NNODES ;

  blockSizeMax = (N + NNODES - 1) / NNODES ;
  blockBase    = blockSizeMax * node_id ;

  if(blockBase + blockSizeMax > N)
    blockSize = N - blockBase ;
  else
    blockSize = blockSizeMax ;

  /* `block' has `blockSize + 2' columns and is offset so that
     subscripts start from -1.  This allows for ghost cells.  */

  block = (int (*) [N]) malloc((blockSize + 2) * sizeof(int [N])) + 1 ;

  /*  Initialise board.  */

  for(i = 0 ; i < blockSize ; i++) {
    for(y = 0 ; y < N ; y++) {
      int x = blockBase + i ;
      if(x == N / 2 || y == N / 2)
        block [i] [y] = 1 ;
      else
        block [i] [y] = 0 ;
    }
  }

  /*  Dump initial state of board to host */

  pvm_psend(tid_host, 1, block, blockSize * N, PVM_INT) ;

  /*  Main update loop.  */

  neighbours = (int (*) [N]) malloc(blockSize * sizeof(int [N])) ;

  for(iter = 0 ; iter < NITER ; iter++) {

    /*  Shift this block's upper edge into next neighbour's lower ghost edge */

    pvm_psend(tids [node_next], 0, block [blockSize - 1], N, PVM_INT) ;
    pvm_precv(tids [node_prev], 0, block [-1],            N, PVM_INT,
              &dummy, &dummy, &dummy) ;

    /*  Shift this block's lower edge into prev neighbour's upper ghost edge */

    pvm_psend(tids [node_prev], 0, block [0],         N, PVM_INT) ;
    pvm_precv(tids [node_next], 0, block [blockSize], N, PVM_INT,
              &dummy, &dummy, &dummy) ;

    /*  Calculate a block of neighbour sums.  */

    for(i = 0 ; i < blockSize ; i++) {
      for(y = 0 ; y < N ; y++) {
        int y_n = (y - 1 + N) % N ;
        int y_p = (y + 1) % N ;

        neighbours [i] [y] =
             block [i - 1] [y_n] + block [i - 1] [y] + block [i - 1] [y_p] +
             block [i]     [y_n] +                     block [i]     [y_p] +
             block [i + 1] [y_n] + block [i + 1] [y] + block [i + 1] [y_p] ;
      }
    }

    /*  Update block of board values.  */

    for(i = 0 ; i < blockSize ; i++) {
      for(y = 0 ; y < N ; y++) {
        int neighbour = neighbours [i] [y] ;

        if(neighbour < 2 || neighbour > 3)
          block [i] [y] = 0 ;
        if(neighbour == 3)
          block [i] [y] = 1 ;
      }
    }

    /*  Dump current state of board to host */

    pvm_psend(tid_host, 1, block, blockSize * N, PVM_INT) ;
  }

  pvm_exit() ;
}

